@@ -18,6 +18,12 @@ class Message < Retriever::Model
   # screen nameのみから構成される文字列から、@などを切り取るための正規表現
   MentionExactMatcher = /\A(?:@|＠|〄|☯|⑨|♨|D )?([a-zA-Z0-9_]+)\Z/.freeze
 
+  PermalinkMatcher = Regexp.union(
+    %r[\Ahttps?://twitter.com/(?:#!/)?(?<screen_name>[a-zA-Z0-9_]+)/status(?:es)?/(?<id>\d+)(?:\?.*)?\Z], # Twitter
+    %r[\Ahttp://favstar\.fm/users/(?<screen_name>[a-zA-Z0-9_]+)/status/(?<id>\d+)], # Hey, Favstar. Ban stop me premiamu!
+    %r[\Ahttp://aclog\.koba789\.com/i/(?<id>\d+)] # Hey, Twitter. Please BAN me rhenium!
+  ).freeze
+
   extend Gem::Deprecate
 
   @@system_id = 0
@@ -129,7 +135,7 @@ class Message < Retriever::Model
 
   # この投稿をリツイートする権限があればtrueを返す
   def retweetable?
-    Service.primary and not system? and not from_me? and not user[:protected] end
+    Service.primary and not system? and not from_me? and not protected? end
 
   # この投稿を削除する権限があればtrueを返す
   def deletable?
@@ -144,12 +150,18 @@ class Message < Retriever::Model
   def to_me?
     system? or Service.map(&:user_obj).find(&method(:receive_to?)) end
 
+  # この投稿が公開されているものならtrueを返す。少しでも公開範囲を限定しているならfalseを返す。
+  def protected?
+    user.protected? end
+
+  # この投稿が承認されているものならtrueを返す。
+  def verified?
+    user.verified? end
+
   # この投稿の投稿主を返す
   def user
     self.get(:user, -1) end
 
-  # この投稿のServiceオブジェクトを返す。
-  # 設定されてなければnilを返す
   def service
     warn "Message#service is obsolete method. use `Service.primary'."
     Service.primary end
@@ -187,7 +199,8 @@ class Message < Retriever::Model
 
   # このメッセージが何かしらの別のメッセージに宛てられたものなら真
   def has_receive_message?
-    self[:replyto] end
+    !!self[:replyto] end
+  alias reply? has_receive_message?
 
   # このメッセージが何かに対するリツイートなら真
   def retweet?
@@ -204,48 +217,245 @@ class Message < Retriever::Model
   def receive_message_d(force_retrieve=false)
     Thread.new{ receive_message(force_retrieve) } end
 
-  def self.define_source_getter(key, condition=ret_nth, &onfound)
-    define_method("#{key}_source"){ |*args|
-      force_retrieve = args.first
-      if(condition === self[:message].to_s)
-        result = get(key, (force_retrieve ? -1 : 1))
-        if result.is_a?(Message)
-          onfound.call(self, result)
-          result end end }
-    define_method("#{key}_source_d"){ |*args|
-      Thread.new{ __send__("#{key}_source", *args) } } end
+  # このMessageの宛先になっているMessageを取得して返す。
+  # ==== Args
+  # [force_retrieve] 真なら、ツイートがメモリ上に見つからなかった場合Twitter APIリクエストを発行する
+  # ==== Return
+  # Message|nil 宛先のMessage。宛先がなければnil
+  def replyto_source(force_retrieve=false)
+    if reply?
+      result = get(:replyto, (force_retrieve ? -1 : 1))
+      if result.is_a?(Message)
+        result.add_child(self) unless result.children.include?(self)
+        result end end end
 
-  # Message#receive_message と同じ。ただし、リプライ元のみをさがす。
-  define_source_getter(:replyto, /@[a-zA-Z0-9_]/){ |this, result|
-    result.add_child(this) unless result.children.include?(this) }
+  # replyto_source の戻り値をnextに渡すDeferredableを返す
+  # ==== Args
+  # [force_retrieve] 真なら、ツイートがメモリ上に見つからなかった場合Twitter APIリクエストを発行する
+  # ==== Return
+  # Deferredable nextの引数に宛先のMessageを渡す。宛先が無い場合は失敗し、trap{}にnilを渡す
+  def replyto_source_d(force_retrieve=true)
+    Thread.new do
+      result = replyto_source(force_retrieve)
+      if result.is_a? Message
+        result
+      else
+        Deferred.fail(result) end end end
 
-  # Message#receive_message と同じ。ただし、ReTweetedのみをさがす。
-  define_source_getter(:retweet, -> t { t.start_with? 'RT'.freeze }){ |this, result|
-    result.add_child(this) unless result.retweeted_statuses.include?(this) }
+  # このMessageがリツイートであるなら、リツイート元のツイートを返す。
+  # リツイートではないならnilを返す。リツイートであるかどうかを確認するには、
+  # このメソッドの代わりに Message#retweet? を使う。
+  # ==== Args
+  # [force_retrieve]
+  # ==== Return
+  # Message|nil リツイート元のMessage。リツイートではないならnil
+  def retweet_parent(force_retrieve=false)
+    if retweet?
+      result = get(:retweet, (force_retrieve ? -1 : 1))
+      if result.is_a?(Message)
+        result.add_child(self) unless result.retweeted_statuses.include?(self)
+        result end end end
+
+  # retweet_parent の戻り値をnextに渡すDeferredableを返す
+  # ==== Args
+  # [force_retrieve] 真なら、ツイートがメモリ上に見つからなかった場合Twitter APIリクエストを発行する
+  # ==== Return
+  # Deferredable nextの引数にリプライ元のMessageを渡す。リツイートではない場合は失敗し、trap{}にnilを渡す
+  def retweet_parent_d(force_retrieve=true)
+    Thread.new do
+      result = retweet_source(force_retrieve)
+      if result.is_a? Message
+        result
+      else
+        Deferred.fail(result) end end end
+
+  # このMessageが引用した投稿を全て返す
+  # ==== Return
+  # Enumerable このMessageが引用したMessageのid(Fixnum)
+  def quoting_ids
+    entity.lazy.select{ |entity|
+      :urls == entity[:slug]
+    }.map{ |entity|
+      PermalinkMatcher.match(entity[:expanded_url])
+    }.select(&ret_nth).map do |matched|
+      matched[:id].to_i end end
+
+  # このMessageが引用した投稿を全て返す。
+  # _force_retrieve_ に真が指定されたらこのメソッドはTwitter APIをリクエストする可能性がある。
+  # そのため _force_retrieve_ が真なら、Messageを取得してから返し、
+  # 偽ならAPIリクエストが必要ないので、Messageオブジェクトの取得を遅延する。
+  # Twitter APIリクエストを行ったがツイートが削除されていた、メモリ上に存在しないなどの理由で
+  # 取得できなかったツイートに関しては、戻り値に含まれない
+  # ==== Args
+  # [force_retrieve] 真なら、ツイートがメモリ上に見つからなかった場合Twitter APIリクエストを発行する
+  # ==== Return
+  # Enumerable このMessageが引用したMessage
+  def quoting_messages(force_retrieve=false)
+    return @quoting_messages if defined? @quoting_messages
+    if force_retrieve
+      @quoting_messages ||= quoting_ids.map{|quoted_id|
+        Message.findbyid(quoted_id, -1)
+      }.to_a.compact.freeze.tap do |qs|
+        qs.each do |q|
+          q.add_quoted_by(self) end  end
+    else
+      quoting_ids.map{|quoted_id|
+        Message.findbyid(quoted_id, 0) }.select(&ret_nth) end end
+
+  # このMessageが引用した投稿を全て返す。
+  # _force_retrieve_ に真が指定されたらこのメソッドはTwitter APIをリクエストする可能性がある。
+  # Twitter APIリクエストを行ったがツイートが削除されていた、メモリ上に存在しないなどの理由で
+  # 取得できなかったツイートに関しては、結果がnilとなる
+  # ==== Args
+  # [force_retrieve] 真なら、ツイートがメモリ上に見つからなかった場合Twitter APIリクエストを発行する
+  # ==== Return
+  # Deferredable
+  def quoting_messages_d(force_retrieve=false)
+    Thread.new{ quoting_messages(force_retrieve) } end
+
+  # self が、何らかのツイートを引用しているなら真を返す
+  # ==== Return
+  # TrueClass|FalseClass
+  def quoting?
+    !!quoting_ids.first end
+
+  # selfを引用しているツイート _message_ を登録する
+  # ==== Args
+  # [message] Message selfを引用しているMessage
+  # ==== Return
+  # self
+  def add_quoted_by(message)
+    atomic do
+      @quoted_by ||= Messages.new
+      unless @quoted_by.include? message
+        if @quoted_by.frozen?
+          @quoted_by = Messages.new(@quoted_by + [message])
+        else
+          @quoted_by << message end end
+      self end end
+
+  # selfを引用しているツイートを返す
+  # ==== Return
+  # Messages selfを引用しているMessageの配列
+  def quoted_by
+    if defined? @quoted_by
+      @quoted_by
+    else
+      atomic do
+        @quoted_by ||= Messages.new end end.freeze end
+
+  # self が、何らかのツイートから引用されているなら真を返す
+  # ==== Return
+  # TrueClass|FalseClass
+  def quoted_by?
+    !quoted_by.empty? end
+
+  # 投稿の宛先になっている投稿を再帰的にさかのぼるような _Enumerator_ を返す。
+  # ==== Return
+  # Enumerator
+  def ancestors_enumerator(force_retrieve=false)
+    Enumerator.new do |yielder|
+      message = self
+      while message
+        yielder << message
+        message = message.receive_message(force_retrieve) end end end
+  private :ancestors_enumerator
 
   # 投稿の宛先になっている投稿を再帰的にさかのぼり、それぞれを引数に取って
   # ブロックが呼ばれる。
+  # ブロックが渡されていない場合、 _Enumerator_ を返す。
   # _force_retrieve_ は、 Message#receive_message の引数にそのまま渡される
-  def each_ancestors(force_retrieve=false, &proc)
-    proc.call(self)
-    parent = receive_message(force_retrieve)
-    parent.each_ancestors(force_retrieve, &proc) if parent
-  end
+  # ==== Return
+  # obj|Enumerator
+  def each_ancestor(force_retrieve=false, &proc)
+    e = ancestors_enumerator(force_retrieve)
+    if block_given?
+      e.each(&proc)
+    else
+      e end end
+  alias :each_ancestors :each_ancestor
+  deprecate :each_ancestors, "each_ancestor", 2016, 12
 
   # 投稿の宛先になっている投稿を再帰的にさかのぼり、それらを配列にして返す。
   # 配列インデックスが大きいものほど、早く投稿された投稿になる。
   # （[0]は[1]へのリプライ）
   def ancestors(force_retrieve=false)
-    parent = receive_message(force_retrieve)
-    return [self, *parent.ancestors(force_retrieve)] if parent
-    [self] end
-  memoize :ancestors
+    ancestors_enumerator(force_retrieve).to_a end
 
   # 投稿の宛先になっている投稿を再帰的にさかのぼり、何にも宛てられていない投稿を返す。
   # つまり、一番祖先を返す。
   def ancestor(force_retrieve=false)
     ancestors(force_retrieve).last end
 
+  # retweet元を再帰的にさかのぼるような _Enumerator_ を返す。
+  # この _Enumerator_ は最初にこの _Message_ 自身を yield し、以降は直前に yield した
+  # 要素のretweet元を yield する。
+  # ==== Return
+  # Enumerator
+  def retweet_ancestors_enumerator(force_retrieve=false)
+    Enumerator.new do |yielder|
+      message = self
+      while message
+        yielder << message
+        message = message.retweet_parent(force_retrieve)
+      end end end
+  private :retweet_ancestors_enumerator
+
+  # retweet元を再帰的にさかのぼり、それぞれを引数に取って
+  # ブロックが呼ばれる。
+  # ブロックが渡されていない場合、 _Enumerator_ を返す。
+  # _force_retrieve_ は、 Message#retweet_parent の引数にそのまま渡される
+  # ==== Return
+  # obj|Enumerator
+  def each_retweet_ancestor(force_retrieve=false, &proc)
+    e = retweet_ancestors_enumerator(force_retrieve)
+    if block_given?
+      e.each(&proc)
+    else
+      e end end
+  alias :each_retweet_ancestors :each_retweet_ancestor
+  deprecate :each_retweet_ancestors, "each_retweet_ancestor", 2016, 12
+
+  # retweet元を再帰的に遡り、それらを配列にして返す。
+  # 配列の最初の要素は必ずselfになり、以降は直前の要素のリツイート元となる。
+  # ([0]は[1]へのリツイート)
+  # ==== Return
+  # Enumerator
+  def retweet_ancestors(force_retrieve=false)
+    retweet_ancestors_enumerator(force_retrieve).to_a end
+
+  # リツイート元を再帰的に遡り、リツイートではないツイートを返す。
+  # selfがリツイートでない場合は、selfを返す。
+  # ==== Args
+  # [force_retrieve] 真なら、ツイートがメモリ上に見つからなかった場合Twitter APIリクエストを発行する
+  # ==== Return
+  # Message
+  def retweet_ancestor(force_retrieve=false)
+    retweet_ancestors(force_retrieve).last end
+
+  # このMessageがリツイートなら、何のリツイートであるかを返す。
+  # 返される値の retweet? は常に false になる
+  # ==== Args
+  # [force_retrieve] 真なら、ツイートがメモリ上に見つからなかった場合Twitter APIリクエストを発行する
+  # ==== Return
+  # Message|nil リツイートであればリツイート元のMessage、リツイートでなければnil
+  def retweet_source(force_retrieve=false)
+    if retweet?
+      retweet_ancestor(force_retrieve) end end
+
+  # retweet_source の戻り値をnextに渡すDeferredableを返す
+  # ==== Args
+  # [force_retrieve] 真なら、ツイートがメモリ上に見つからなかった場合Twitter APIリクエストを発行する
+  # ==== Return
+  # Deferredable nextの引数にリプライ元のMessageを渡す。リツイートではない場合は失敗し、trap{}にnilを渡す
+  def retweet_source_d(force_retrieve=true)
+    Thread.new do
+      result = retweet_source(force_retrieve)
+      if result.is_a? Message
+        result
+      else
+        Deferred.fail(result) end end end
+
   # このMessageが属する親子ツリーに属する全てのMessageを含むSetを返す
   # ==== Args
   # [force_retrieve] 外部サーバに問い合わせる場合真
@@ -256,7 +466,7 @@ class Message < Retriever::Model
 
   # この投稿に宛てられた投稿をSetオブジェクトにまとめて返す。
   def children
-    @children ||= Plugin.filtering(:replied_by, self, Set.new())[1] + retweeted_statuses end
+    @children ||= Plugin.filtering(:replied_by, self, Set.new(retweeted_statuses))[1] end
 
   # childrenを再帰的に遡り全てのMessageを返す
   # ==== Return
@@ -280,16 +490,20 @@ class Message < Retriever::Model
 
   # この投稿をリツイートしたユーザを返す
   def retweeted_by
-    retweeted_statuses.map{ |x| x.user }.uniq
-  end
+    retweeted_sources.lazy.map(&:user) end
+  alias retweeted_users retweeted_by
 
   # この投稿に対するリツイートを返す
   def retweeted_statuses
-    @retweets ||= Plugin.filtering(:retweeted_by, self, Set.new)[1].to_a.compact end
+    retweeted_sources.lazy.select{|m| m.is_a?(Message) } end
+
+  # この投稿に対するリツイートまたはユーザを返す
+  def retweeted_sources
+    @retweets ||= Plugin.filtering(:retweeted_by, self, Set.new())[1].to_a.compact end
 
   # 選択されているユーザがこのツイートをリツイートしているなら真
   def retweeted?
-    retweeted_by.include?(Service.primary!.user_obj)
+    retweeted_users.include?(Service.primary!.user_obj)
   rescue Service::NotExistError
     false end
 
@@ -297,12 +511,23 @@ class Message < Retriever::Model
   def retweeted_by_me?(me = Service.services)
     case me
     when Service
-      retweeted_by.include? me.user_obj
+      retweeted_users.include? me.user_obj
     when Enumerable
-      not (Set.new(retweeted_by.map(&:idname)) & Set.new(me.map(&:idname))).empty?
+      not (Set.new(retweeted_users.map(&:idname)) & Set.new(me.map(&:idname))).empty?
     else
       raise ArgumentError, "first argument should be `Service' or `Enumerable'. but given `#{me.class}'" end end
 
+  # この投稿をリツイート等して、 _me_ のタイムラインに出現させたリツイートを返す。
+  # 特に誰もリツイートしていない場合は _self_ を返す。
+  # リツイート、ふぁぼなどを行う時に使用する。
+  # ==== Args
+  # [me] Service 対象とするService
+  # ==== Return
+  # Message
+  def introducer(me = Service.primary!)
+    Plugin.filtering(:message_introducers, me, self, retweeted_statuses.reject{|m|m.user == me.to_user}).last.to_a.last || self
+  end
+
   # 非公式リツイートやハッシュタグを適切に組み合わせて投稿する
   def body
     self[:message].to_s.freeze
@@ -352,6 +577,7 @@ class Message < Retriever::Model
   # :nodoc:
   def add_favorited_by(user, time=Time.now)
     type_strict user => User, time => Time
+    return retweet_source.add_favorited_by(user, time) if retweet?
     service = Service.primary
     if service
       set_modified(time) if UserConfig[:favorited_by_anyone_age] and (UserConfig[:favorited_by_myself_age] or service.user != user.idname)
@@ -361,6 +587,7 @@ class Message < Retriever::Model
   # :nodoc:
   def remove_favorited_by(user)
     type_strict user => User
+    return retweet_source.remove_favorited_by(user) if retweet?
     service = Service.primary
     if service
       favorited_by.delete(user)
@@ -374,7 +601,7 @@ class Message < Retriever::Model
         add_retweet_in_this_thread(child)
       else
         SerialThread.new{
-          retweeted_by
+          retweeted_sources
           add_retweet_in_this_thread(child) } end
     else
       if defined? @children
@@ -384,19 +611,39 @@ class Message < Retriever::Model
           children
           add_child_in_this_thread(child) } end end end
 
+  # :nodoc:
+  def add_retweet_user(retweet_user, created_at)
+    type_strict retweet_user => User
+    return retweet_source.add_retweet_user(retweet_user, created_at) if retweet?
+    if defined? @retweets
+      add_retweet_in_this_thread(retweet_user, created_at)
+    else
+      SerialThread.new{
+        retweeted_sources
+        add_retweet_in_this_thread(retweet_user, created_at) } end end
+
   # 最終更新日時を取得する
   def modified
-    @value[:modified] ||= [self[:created], *(defined?(@retweets) ? @retweets : []).map{ |x| x.modified }].compact.max
+    @value[:modified] ||= [self[:created], *(@retweets || []).map{ |x| x.modified }].compact.max
+  end
+
+  def inspect
+    "#<#{self.class.name}: #{id} #{user.inspect} #{to_show}>"
   end
 
   private
 
-  def add_retweet_in_this_thread(child)
-    type_strict child => Message
-    @retweets = [] if not defined? @retweets
-    @retweets << child
+  def add_retweet_in_this_thread(child, created_at=child[:created])
+    type_strict child => tcor(Message, User)
+    unless @retweets.include? child
+      case child
+      when Message
+        @retweets << child
+        @retweets.delete(child.user) if @retweets.include?(child.user)
+      when User
+        @retweets << child if retweeted_users.include?(child) end end
     service = Service.primary
-    set_modified(child[:created]) if service and UserConfig[:retweeted_by_anyone_age] and ((UserConfig[:retweeted_by_myself_age] or service.user != child.user.idname)) end
+    set_modified(created_at) if service and UserConfig[:retweeted_by_anyone_age] and ((UserConfig[:retweeted_by_myself_age] or service.user != child.user.idname)) end
 
   def add_child_in_this_thread(child)
     @children << child
